from __future__ import annotations

import asyncio
import dataclasses
import shutil
from typing import (
    ClassVar,
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
)

from slurmbridge.cliobject import SlurmCLIObject, SlurmObjectException
from slurmbridge.common import camel_to_snake_case, snake_to_camel_case

SCONTROL_PATH = shutil.which("scontrol")


class SlurmControlError(SlurmObjectException):
    def __str__(self) -> str:
        return f"{self.args[1]}"


SlurmControlObjectType = TypeVar("SlurmControlObjectType", bound="SlurmControlObject")


class SlurmControlObject(SlurmCLIObject[SlurmControlObjectType]):
    query_options: ClassVar[Sequence[str]] = tuple()
    query_type: ClassVar[Literal["Normal", "ValueOnly"]] = "Normal"

    @classmethod
    async def _run_scontrol_command(
        cls,
        verb: str,
        *arguments: str,
        error_ok=False,
    ) -> Tuple[int | None, str]:
        if SCONTROL_PATH is None:
            raise ImportError("scontrol could not be found in path.")

        process = await asyncio.create_subprocess_exec(
            SCONTROL_PATH,
            verb,
            *arguments,
            "--oneline",
            "--detail",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await process.communicate()

        process_output = (stderr if len(stderr) > 0 else stdout).decode("ascii")

        if not error_ok and process.returncode != 0:
            raise SlurmControlError(cls, process_output)

        return process.returncode, process_output

    @classmethod
    async def _scontrol_show(
        cls,
        fields: Sequence[str],
        filters: Dict[str, str],
    ) -> Sequence[Dict[str, str]]:
        filter_args: list[str] = []

        # in scontrol you pass the object type to get all entries of an object
        # OR you pass a primary key to filter for specific instances but you can't do both
        if len(filters) > 0:
            if cls.query_type == Literal["ValueOnly"]:
                filter_args.append(cls.__name__)
                filter_args.append(f"{list(filters.values())[0]}")
            else:
                for column, value in filters.items():
                    filter_args.append(f"{column}={value}")
        else:
            filter_args.append(cls.__name__)

        exit_code, process_output = await cls._run_scontrol_command(
            "show", *filter_args, *cls.query_options
        )
        if exit_code != 0:
            raise SlurmControlError(cls, process_output)

        all_objects = []
        case_insensitive_fields = [field_name.lower() for field_name in fields]
        # we use --oneline to make sure we have exactly 1 object per line
        for object_reponse in process_output.splitlines():
            # attributes are seperated by any whitespace
            object_attributes = {}
            for attribute_string in object_reponse.split():
                # sometimes an attribute is not followed by a value (e.g. "Name=")
                # so we pack second argument to a list but thanks to maxsplit=1
                # this list is either empty or has exactly one entry
                attribute_name, *attribute_value = attribute_string.split(
                    "=", maxsplit=1
                )
                # we need to use case insensitive matching here since slurm
                # uses SREAMINGCASE for abbrevations (e.g. TRES instead of Tres)
                if attribute_name.lower() in case_insensitive_fields:
                    object_attributes[attribute_name] = (
                        attribute_value[0] if len(attribute_value) == 1 else ""
                    )
            if len(object_attributes) > 0:
                all_objects.append(object_attributes)

        return all_objects

    @classmethod
    async def _scontrol_execute(
        cls, verb: str, filter: str, extra_args: Optional[List[str]] = None
    ):
        if extra_args is None:
            extra_args = []
        exit_code, process_output = await cls._run_scontrol_command(
            verb,
            *extra_args,
            filter,
            error_ok=True,
        )

        if exit_code != 0:
            raise SlurmControlError(cls, process_output)

    @classmethod
    async def _scontrol_update(
        cls,
        new_values: Mapping[str, str],
        filters: Mapping[str, str],
    ):
        update_args: list[str] = []
        for field_name, new_value in new_values.items():
            update_args.append(f"{field_name}={new_value}")

        filter_args = []
        for column, value in filters.items():
            filter_args.append(f"{column}={value}")

        exit_code, process_output = await cls._run_scontrol_command(
            "update",
            *filter_args,
            *update_args,
            error_ok=True,
        )

        if exit_code != 0:
            raise SlurmControlError(cls, process_output)

    @classmethod
    def _response_to_attributes(
        cls: Type[SlurmControlObjectType], response_data: Dict[str, str]
    ) -> Dict[str, str | None]:
        empty_variants = ("(null)", "N/A", "None")
        return {
            # some values will always return a value although they are empty
            # we need to ignore them here to make sure not to set these values
            # when saving the object again
            camel_to_snake_case(key): value if value not in empty_variants else None
            for key, value in response_data.items()
        }

    @classmethod
    def _response_to_instances(
        cls: Type[SlurmControlObjectType], response: Sequence[Dict[str, str]]
    ) -> Sequence[SlurmControlObjectType]:
        instances: list[SlurmControlObjectType] = []
        for data in response:
            attributes = cls._response_to_attributes(data)
            instance = cls(**attributes)
            instances.append(instance)
        return instances

    def _to_query(self) -> Tuple[Dict[str, str], Dict[str, str]]:
        update_fields: Dict[str, str] = {}
        filter_values = {}

        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if field.name in self._primary_key_fields:
                filter_values[snake_to_camel_case(field.name)] = value
            elif (
                field.name not in (self._read_only_fields + self._synthetic_fields)
                and value
            ):
                update_fields[snake_to_camel_case(field.name)] = value

        return update_fields, filter_values

    @classmethod
    async def filter(
        cls: Type[SlurmControlObjectType], **filters: str
    ) -> Sequence[SlurmControlObjectType]:
        object_fields = [
            snake_to_camel_case(field.name)
            for field in dataclasses.fields(cls)
            if field.name not in (cls._synthetic_fields + cls._write_only_fields)
        ]
        object_data = await cls._scontrol_show(object_fields, filters)
        instances = cls._response_to_instances(object_data)
        # only return the queried type of instance
        return [instance for instance in instances if isinstance(instance, cls)]

    async def save(self):
        updates, filters = self._to_query()

        await self._scontrol_update(updates, filters)
        await self.refresh_from_db()



import dataclasses
from abc import ABC, abstractmethod
from typing import Any, ClassVar, Dict, Generic, List, Sequence, Tuple, Type, TypeVar

READONLY = {"readonly": True}
WRITEONLY = {"writeonly": True}
SYNTHETIC = {"synthetic": True}
PRIMARYKEY = {"primarykey": True}


class SlurmObjectException(Exception):
    def __init__(self, object_class, *args):
        self.object_class = object_class
        super().__init__(object_class, *args)

    def __str__(self) -> str:
        return f"{self.object_class.__name__} - {self.__class__.__name__}"


class NotFound(SlurmObjectException):
    pass


class MultipleObjectReturned(SlurmObjectException):
    pass


def field(
    *,
    default=dataclasses.MISSING,
    default_factory=dataclasses.MISSING,
    init=True,
    repr=True,
    hash=None,
    compare=True,
    metadata=None,
    read_only=False,
    synthetic=False,
    write_only=False,
    primary_key=False,
):
    if metadata is None:
        metadata = {}
    if read_only:
        metadata |= READONLY
    if synthetic:
        metadata |= SYNTHETIC | READONLY
    if primary_key:
        metadata |= PRIMARYKEY | READONLY
    if write_only:
        metadata |= WRITEONLY

    return dataclasses.field(
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


ObjectType = TypeVar("ObjectType", bound="SlurmCLIObject")


class SlurmCLIObject(ABC, Generic[ObjectType]):
    _primary_key_fields: ClassVar[List[str]]
    _read_only_fields: ClassVar[List[str]]
    _write_only_fields: ClassVar[List[str]]
    _synthetic_fields: ClassVar[List[str]]

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        cls = dataclasses.dataclass(cls)
        cls._primary_key_fields = cls._collect_all_fields_of_type(PRIMARYKEY)
        cls._read_only_fields = cls._collect_all_fields_of_type(READONLY)
        cls._write_only_fields = cls._collect_all_fields_of_type(WRITEONLY)
        cls._synthetic_fields = cls._collect_all_fields_of_type(SYNTHETIC)

    @classmethod
    def _collect_all_fields_of_type(cls, field_type: Dict[str, bool]):
        type_key = list(field_type.keys())[0]
        return [
            field.name
            for field in dataclasses.fields(cls)
            if field.metadata is not None and field.metadata.get(type_key, False)
        ]

    @classmethod
    async def all(cls) -> Sequence[ObjectType]:
        return await cls.filter()

    @classmethod
    async def get(cls, **filters: str) -> ObjectType:
        object_list = await cls.filter(**filters)
        if len(object_list) == 0:
            raise NotFound(cls)
        if len(object_list) > 1:
            raise MultipleObjectReturned(cls)
        return object_list[0]

    async def refresh_from_db(self):
        _, filters = self._to_query()
        new_object = await self.get(**filters)
        for field in dataclasses.fields(new_object):
            del self.__dict__[field.name]
            setattr(self, field.name, getattr(new_object, field.name))

    @classmethod
    @abstractmethod
    async def filter(cls: Type[ObjectType], **filters: str) -> Sequence[ObjectType]:
        ...

    @abstractmethod
    def _to_query(self) -> Tuple[Dict[str, str], Dict[str, str]]:
        ...



from __future__ import annotations

import asyncio
import dataclasses
import shutil
from typing import ClassVar, Dict, Mapping, Optional, Sequence, Tuple, Type, TypeVar

from slurmbridge.cliobject import SlurmCLIObject, SlurmObjectException
from slurmbridge.common import camel_to_snake_case, snake_to_camel_case

SACCTMGR_PATH = shutil.which("sacctmgr")


class SlurmAccountManagerError(SlurmObjectException):
    def __str__(self) -> str:
        return f"{self.args[1]}"


AccountManagerObjectType = TypeVar(
    "AccountManagerObjectType", bound="SlurmAccountManagerObject"
)


class SlurmAccountManagerObject(SlurmCLIObject[AccountManagerObjectType]):
    query_options: ClassVar[Sequence[str]] = tuple()

    @classmethod
    async def _run_scattmgr_command(
        cls,
        verb: str,
        *arguments: str,
        error_ok=False,
    ) -> Tuple[int | None, str]:
        if SACCTMGR_PATH is None:
            raise ImportError("sacctmgr could not be found in path.")

        process = await asyncio.create_subprocess_exec(
            SACCTMGR_PATH,
            verb,
            *arguments,
            "--parsable2",  # seperate values by a PIPE
            "--noheader",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await process.communicate()

        process_output = (stderr if len(stderr) > 0 else stdout).decode("ascii")

        if not error_ok and process.returncode != 0:
            raise SlurmAccountManagerError(cls, process_output)

        return process.returncode, process_output

    @classmethod
    async def _scattmgr_show(
        cls,
        fields: Sequence[str],
        filters: Optional[Mapping[str, str]] = None,
    ) -> Sequence[Dict[str, str]]:
        format_string = "format=" + ",".join(fields)

        filter_args: list[str] = []
        if filters is not None and len(filters) > 0:
            filter_args.append("where")
            for column, value in filters.items():
                filter_args.append(f"{column}={value}")

        exit_code, process_output = await cls._run_scattmgr_command(
            "show", cls.__name__, *cls.query_options, format_string, *filter_args
        )

        return [
            dict(zip(fields, line.split("|"))) for line in process_output.splitlines()
        ]

    @classmethod
    def _response_to_attributes(
        cls: Type[AccountManagerObjectType], response_data: Dict[str, str]
    ) -> Dict[str, str | None]:
        empty_variants = ("0-00:00:00", "", "0", "00:00:00")
        return {
            # some values will always return a value although they are empty
            # we need to ignore them here to make sure not to set these values
            # when saving the object again
            camel_to_snake_case(key): value if value not in empty_variants else None
            for key, value in response_data.items()
        }

    @classmethod
    def _response_to_instances(
        cls: Type[AccountManagerObjectType], response: Sequence[Dict[str, str]]
    ) -> Sequence[AccountManagerObjectType]:
        instances: list[AccountManagerObjectType] = []
        for data in response:
            attributes = cls._response_to_attributes(data)
            instance = cls(**attributes)
            instances.append(instance)
        return instances

    def _to_query(self) -> Tuple[Dict[str, str], Dict[str, str]]:
        update_fields: Dict[str, str] = {}
        filter_fields: Dict[str, str] = {}

        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if field.name in self._primary_key_fields:
                filter_fields[snake_to_camel_case(field.name)] = value
            elif (
                field.name not in (self._read_only_fields + self._synthetic_fields)
                and value
            ):
                update_fields[snake_to_camel_case(field.name)] = value

        return update_fields, filter_fields

    @classmethod
    async def filter(
        cls: Type[AccountManagerObjectType], **filters: str
    ) -> Sequence[AccountManagerObjectType]:
        object_fields = [
            snake_to_camel_case(field.name)
            for field in dataclasses.fields(cls)
            if field.name not in (cls._synthetic_fields + cls._write_only_fields)
        ]
        object_data = await cls._scattmgr_show(
            object_fields,
            filters,
        )
        instances = cls._response_to_instances(object_data)
        # only return the queried type of instance
        return [instance for instance in instances if isinstance(instance, cls)]


WritableAccountManagerObjectType = TypeVar(
    "WritableAccountManagerObjectType", bound="WritableSlurmAccountManagerObject"
)


class WritableSlurmAccountManagerObject(
    SlurmAccountManagerObject[WritableAccountManagerObjectType]
):
    @classmethod
    async def _scattmgr_write(
        cls,
        verb: str,
        new_values: Mapping[str, str],
        filters: Mapping[str, str],
    ) -> Sequence[str]:
        filter_args: list[str] = []
        if len(filters) > 0:
            filter_args.append("where")
            for column, value in filters.items():
                filter_args.append(f"{column}={value}")

        update_args: list[str] = []
        if len(new_values) > 0:
            if verb == "modify":
                update_args.append("set")
            for field_name, new_value in new_values.items():
                update_args.append(f"{field_name}={new_value}")

        exit_code, process_output = await cls._run_scattmgr_command(
            verb,
            cls.__name__,
            *filter_args,
            *update_args,
            "--immediate",
            error_ok=True,
        )

        if exit_code != 0:
            raise SlurmAccountManagerError(cls, process_output)

        if verb == "create":
            if "Nothing new added." in process_output:
                return []
            else:
                return [""]

        modified_objects = []
        # the processoutput is grouped into one or more sections where
        # the first section always lists the modified objects and other
        # sections may list related modifications
        section = 0
        for line in process_output.splitlines():
            # section headers are formatted like: "Modified <objecttype>..."
            if line.endswith("..."):
                section += 1
                continue
            if section == 1:
                modified_objects.append(line.strip())
            else:
                break

        return modified_objects

    @classmethod
    async def create(
        cls: Type[WritableAccountManagerObjectType], **attrs
    ) -> WritableAccountManagerObjectType:
        created_objects = await cls._scattmgr_write("create", attrs, {})
        if len(created_objects) == 1:
            return await cls.get(**attrs)
        else:
            raise SlurmAccountManagerError(
                cls,
                f"Failed to create new {cls.__name__}. Maybe the object already existed?",
            )

    async def save(self, allow_multiple_affected=False) -> bool:
        created = False
        updates, filters = self._to_query()

        updated_keys = await self._scattmgr_write("modify", updates, filters)
        if len(updated_keys) == 0:
            # the object was not yet present in the db, create a new one
            await self._scattmgr_write("create", updates | filters, {})
            created = True
        elif len(updated_keys) > 1 and not allow_multiple_affected:
            raise SlurmAccountManagerError(
                f"Modified more than a single Object!. Modified keys: {updated_keys}"
            )

        await self.refresh_from_db()

        return created

    async def delete(self) -> bool:
        _, filters = self._to_query()

        updated_keys = await self._scattmgr_write("delete", {}, filters)
        if len(updated_keys) > 1:
            raise SlurmAccountManagerError(
                f"Deleted more than a single Object!. Deleted keys: {updated_keys}"
            )

        return len(updated_keys) == 1



from __future__ import annotations

import asyncio
import shutil
from typing import Optional, Tuple

from slurmbridge.cliobject import SlurmObjectException

SCANCEL_PATH = shutil.which("scancel")


class SlurmCancelError(SlurmObjectException):
    def __str__(self) -> str:
        return f"{self.args[1]}"


class SlurmCancelObject:
    @classmethod
    async def _scancel_execute(
        cls, job_id: str, signal: Optional[str] = None
    ) -> Tuple[int | None, str]:
        if SCANCEL_PATH is None:
            raise ImportError("scancel could not be found in path.")

        signal_parameters = []
        if signal is not None:
            signal_parameters = ["--signal", signal]

        process = await asyncio.create_subprocess_exec(
            SCANCEL_PATH,
            *signal_parameters,
            job_id,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await process.communicate()
        process_output = (stderr if len(stderr) > 0 else stdout).decode("ascii")
        if process.returncode != 0:
            raise SlurmCancelError(cls, process_output)

        return process.returncode, process_output
